"""Automatically generated by hassfest.

To update, run python3 -m script.hassfest
"""

# fmt: off

ZEROCONF = {
    "_Volumio._tcp.local.": [
        {
            "domain": "volumio"
        }
    ],
    "_airplay._tcp.local.": [
        {
            "domain": "samsungtv",
            "manufacturer": "samsung*"
        }
    ],
    "_api._udp.local.": [
        {
            "domain": "guardian"
        }
    ],
    "_axis-video._tcp.local.": [
        {
            "domain": "axis",
            "macaddress": "00408C*"
        },
        {
            "domain": "axis",
            "macaddress": "ACCC8E*"
        },
        {
            "domain": "axis",
            "macaddress": "B8A44F*"
        },
        {
            "domain": "doorbird",
            "macaddress": "1CCAE3*"
        }
    ],
    "_bond._tcp.local.": [
        {
            "domain": "bond"
        }
    ],
    "_daap._tcp.local.": [
        {
            "domain": "forked_daapd"
        }
    ],
    "_dkapi._tcp.local.": [
        {
            "domain": "daikin"
        }
    ],
    "_dvl-deviceapi._tcp.local.": [
        {
            "domain": "devolo_home_control"
        }
    ],
    "_easylink._tcp.local.": [
        {
            "domain": "modern_forms",
            "name": "wac*"
        }
    ],
    "_elg._tcp.local.": [
        {
            "domain": "elgato"
        }
    ],
    "_enphase-envoy._tcp.local.": [
        {
            "domain": "enphase_envoy"
        }
    ],
    "_esphomelib._tcp.local.": [
        {
            "domain": "esphome"
        },
        {
            "domain": "zha",
            "name": "tube*"
        }
    ],
    "_fbx-api._tcp.local.": [
        {
            "domain": "freebox"
        }
    ],
    "_googlecast._tcp.local.": [
        {
            "domain": "cast"
        }
    ],
    "_hap._tcp.local.": [
        {
            "domain": "homekit_controller"
        }
    ],
    "_homekit._tcp.local.": [
        {
            "domain": "homekit"
        }
    ],
    "_http._tcp.local.": [
        {
            "domain": "bosch_shc",
            "name": "bosch shc*"
        },
        {
            "domain": "nam",
            "name": "nam-*"
        },
        {
            "domain": "nam",
            "manufacturer": "nettigo"
        },
        {
            "domain": "rachio",
            "name": "rachio*"
        },
        {
            "domain": "rainmachine",
            "name": "rainmachine*"
        },
        {
            "domain": "shelly",
            "name": "shelly*"
        }
    ],
    "_hue._tcp.local.": [
        {
            "domain": "hue"
        }
    ],
    "_ipp._tcp.local.": [
        {
            "domain": "ipp"
        }
    ],
    "_ipps._tcp.local.": [
        {
            "domain": "ipp"
        }
    ],
    "_kizbox._tcp.local.": [
        {
            "domain": "somfy",
            "name": "gateway*"
        }
    ],
    "_leap._tcp.local.": [
        {
            "domain": "lutron_caseta"
        }
    ],
    "_mediaremotetv._tcp.local.": [
        {
            "domain": "apple_tv"
        }
    ],
    "_miio._udp.local.": [
        {
            "domain": "xiaomi_aqara"
        },
        {
            "domain": "xiaomi_miio"
        },
        {
            "domain": "yeelight",
            "name": "yeelink-*"
        }
    ],
    "_nanoleafapi._tcp.local.": [
        {
            "domain": "nanoleaf"
        }
    ],
    "_nanoleafms._tcp.local.": [
        {
            "domain": "nanoleaf"
        }
    ],
    "_nut._tcp.local.": [
        {
            "domain": "nut"
        }
    ],
    "_plugwise._tcp.local.": [
        {
            "domain": "plugwise"
        }
    ],
    "_powerview._tcp.local.": [
        {
            "domain": "hunterdouglas_powerview"
        }
    ],
    "_printer._tcp.local.": [
        {
            "domain": "brother",
            "name": "brother*"
        }
    ],
    "_sonos._tcp.local.": [
        {
            "domain": "sonos"
        }
    ],
    "_spotify-connect._tcp.local.": [
        {
            "domain": "spotify"
        }
    ],
    "_ssh._tcp.local.": [
        {
            "domain": "smappee",
            "name": "smappee1*"
        },
        {
            "domain": "smappee",
            "name": "smappee2*"
        },
        {
            "domain": "smappee",
            "name": "smappee50*"
        }
    ],
    "_system-bridge._udp.local.": [
        {
            "domain": "system_bridge"
        }
    ],
    "_touch-able._tcp.local.": [
        {
            "domain": "apple_tv"
        }
    ],
    "_viziocast._tcp.local.": [
        {
            "domain": "vizio"
        }
    ],
    "_wled._tcp.local.": [
        {
            "domain": "wled"
        }
    ],
    "_xbmc-jsonrpc-h._tcp.local.": [
        {
            "domain": "kodi"
        }
    ]
}

HOMEKIT = {
    "3810X": "roku",
    "4660X": "roku",
    "7820X": "roku",
    "819LMB": "myq",
    "AC02": "tado",
    "Abode": "abode",
    "BSB002": "hue",
    "C105X": "roku",
    "C135X": "roku",
    "Healty Home Coach": "netatmo",
    "Iota": "abode",
    "LIFX": "lifx",
    "MYQ": "myq",
    "NL*": "nanoleaf",
    "Netatmo Relay": "netatmo",
    "PowerView": "hunterdouglas_powerview",
    "Presence": "netatmo",
    "Rachio": "rachio",
    "SPK5": "rainmachine",
    "Smart Bridge": "lutron_caseta",
    "Socket": "wemo",
    "TRADFRI": "tradfri",
    "Touch HD": "rainmachine",
    "Welcome": "netatmo",
    "Wemo": "wemo",
    "YL*": "yeelight",
    "iSmartGate": "gogogate2",
    "iZone": "izone",
    "tado": "tado"
}
